/*
 * Copyright 2006 Dovetailed Technologies, LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.dovetail.jzos.jbatch;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.util.StringTokenizer;

import org.apache.commons.net.ftp.FTPClient;



import com.dovetail.jzos.FileFactory;

/**
 * Sample program that using the open source
 * <a href="http://jakarta.apache.org/commons/net/">Apache Commons Net</a> FTPClient
 * to download a tab-delimited file containing names and addresses of the US congress and
 * converts it to an XML file.
 * <p>
 * <dt>Arguments to main:</dt>
 * <dd> [outputfile]  [ftpUrl] </dd>
 * <p>
 * <dt>Where:</dt>
 * <dt>outputfile</dt
 * <dd>The xml output file name.  Can be "//dataset.name" or "//DD:ddname" on zOS.
 * defaults to System.out.
 * </dd>
 *
 * <dt>ftpUrl</dt>
 * <dd>defaults to "ftp://ftp.visi.com/users/juan/ContactingCongress.db.txt" 
 * </dd>
 * 
 * <p>
 * This program is portable between workstations and ZOS and can write output XML
 * on ZOS HFS files or //datasets or //DD:ddname.   Writing portable XML files can
 * be tricky since the encoding defaults to "UTF-8" unless specified on the prologue.
 * The JZOS FileFactory is used to open a portable BufferedWriter in a natural 
 * platform encoding and that encoding name is put on the prologue of the XML file.
 * <p>
 * @see FTPClient
 */
public class FtpDownloadToXml {

	private String outputFilename;
	private String outputEncoding;

	private String urlString;

	private FTPClient ftpClient;

	/**
	 * The main method
	 */
	public static void main(String[] args) throws IOException {

		//String urlString = "ftp://ftp.visi.com/users/juan/ContactingCongress.db.txt";
		String urlString = "ftp://209.98.98.2/users/juan/ContactingCongress.db.txt";

		String outputFilename = null;

		if (args.length >= 1) {
			outputFilename = args[0];
		}
		if (args.length >= 2) {
			urlString = args[1];
		}

		FtpDownloadToXml instance = new FtpDownloadToXml(
				urlString, outputFilename);
		instance.run();
	}

	public FtpDownloadToXml(String url, String outputFilename) {

		this.outputFilename = outputFilename;
		this.urlString = url;
	}

	public void run() throws IOException {

		BufferedReader rdr = openReaderOnFtpDownload();
		BufferedWriter wtr = openWriterOnXmlOutput();

		int count = 0;
		
		try {
			String line = rdr.readLine(); // skip the first line which is a header
			writeXmlHeader(wtr);

			while ((line = rdr.readLine()) != null) {
				convertAndWriteXml(line, wtr);
				count++;
			}
			writeXmlTrailer(wtr);
			System.out.println("Downloaded " + count + " records");

		} finally {
			if (ftpClient != null) {
				ftpClient.logout();
				ftpClient.disconnect();
				ftpClient = null;
			}
			if (rdr != null)
				rdr.close();
			if (wtr != null)
				wtr.close();
		}
	}

	protected BufferedReader openReaderOnFtpDownload() throws IOException {
		
		URL url = new URL(urlString);
		File file = new File(url.getFile());
		
		System.out.println("Connecting to: " + url.getHost() + "...");
		
		ftpClient = new FTPClient();
		ftpClient.connect(url.getHost());
		ftpClient.enterLocalPassiveMode();
		
		System.out.println("Connected to: " + url.getHost());

		try {
			if (!ftpClient.login("anonymous", "anonymous@foo.com")) {
				System.err.println("ftp login err:" + ftpClient.getReplyString());
			}

			ftpClient.setFileType(FTPClient.ASCII_FILE_TYPE);
			
			if (!ftpClient.changeWorkingDirectory(file.getParent())) {
				System.err.println("ftp cwd(" + file.getParent() + ") err:" + ftpClient.getReplyString());
			}
			InputStream is = ftpClient.retrieveFileStream(file.getName());
			if (is == null) {
				System.err.println("ftp retreive(" + file.getName() + ") err:" + ftpClient.getReplyString());
			}
			System.out.println("Opened stream on file: " + url);

			// build a buffered reader on the response stream
			// we expect the FTP file to be in ASCII, but we can't be 
			// sure of the current Java file.encoding is, so we excplitly
			// use ISO-8859-1 (latin ascii).
			return new BufferedReader(new InputStreamReader(is, "ISO-8859-1"));

		} catch (IOException ioe) {
			ftpClient.disconnect();
			ftpClient = null;
			throw ioe;
		}
	}

	protected BufferedWriter openWriterOnXmlOutput() throws IOException {
		
		if (outputFilename != null) {
			outputEncoding = FileFactory.getDefaultEncoding(outputFilename);
			return FileFactory.newBufferedWriter(outputFilename, outputEncoding);
		} else {
			outputEncoding = System.getProperty("file.encoding");
			return new BufferedWriter(new OutputStreamWriter(System.out, outputEncoding));
		}

	}
	protected void writeXmlHeader(BufferedWriter wtr) throws IOException {
		wtr.write("<?xml version=\"1.0\" encoding=\"" + outputEncoding + "\" ?>");
		wtr.newLine();
		wtr.write("<congress>");
		wtr.newLine();
	}
	
	protected void writeXmlTrailer(BufferedWriter wtr) throws IOException {
		wtr.write("</congress>");
		wtr.newLine();
	}

	protected void convertAndWriteXml(String line, BufferedWriter wtr)
			throws IOException {

		StringTokenizer strtok = new StringTokenizer(line, "\t");
		
		wtr.write("<member>");
		wtr.newLine();
		
		wtr.write("  <district>");
		wtr.write(strtok.nextToken().trim());
		wtr.write("</district>");
		wtr.newLine();

		wtr.write("  <name>");
		wtr.write(encodeXml(strtok.nextToken().trim()));
		wtr.write("</name>");
		wtr.newLine();

		wtr.write("  <party>");
		wtr.write(strtok.nextToken().trim());
		wtr.write("</party>");
		wtr.newLine();

		wtr.write("  <contact>");
		wtr.newLine();
		
		wtr.write("    <address>");
		wtr.write(encodeXml(strtok.nextToken().trim()));
		wtr.write("</address>");
		wtr.newLine();

		wtr.write("    <phone>");
		wtr.write(strtok.nextToken().trim());
		wtr.write("</phone>");
		wtr.newLine();

		wtr.write("    <district-phone>");
		wtr.write(strtok.nextToken().trim());
		wtr.write("</district-phone>");
		wtr.newLine();

		wtr.write("    <url>");
		wtr.write(encodeXml(strtok.nextToken().trim()));
		wtr.write("</url>");
		wtr.newLine();
		
		wtr.write("  </contact>");
		wtr.newLine();

		wtr.write("</member>");
		wtr.newLine();
		
	}
	
	/**
	 * Encode a String for use as valid XML content by
	 * quoting special XML characters:
	 * <pre>
	 * & -> &amp;
	 * < ->> &lt;
	 * > -> &gt; 
	 * " -> &quot; 
	 * ' -> &apos;
	 * </pre>
	 */
	public static String encodeXml(String s) {
		
		s = replaceAll(s, "&", "&amp;");
		s = replaceAll(s, "<", "&lt;");
		s = replaceAll(s, ">", "&lt;");
		s = replaceAll(s, "\"", "&quot;");
		s = replaceAll(s, "\'", "&apos;");
		return s;
	}
	
	private static String replaceAll(String source, String subStrToReplace, String replaceStr) {
		
		int subStrLen = subStrToReplace.length();
		if (subStrLen == 0) return source;
		int replaceLen = replaceStr.length();
		
		int srcIndex = 0;
		while (srcIndex < source.length()) {
			int i = source.indexOf(subStrToReplace, srcIndex);
			if (i >= 0) {
				source = source.substring(0, i) + replaceStr + source.substring(i + subStrLen);
				srcIndex = i + replaceLen;
			} else {
				break;
			}
		}
		return source;	
	}
}
