/*
 * Copyright 2006 Dovetailed Technologies, LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.dovetail.jzos.jbatch;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;

import org.apache.commons.digester.Digester;

import com.dovetail.jzos.FileFactory;

/**
 * Sample program that using the open source
 * <a href="http://jakarta.apache.org/commons/digester/">Apache Commons Digester</a>
 * to parse an XML file containing names and addresses of the US congress and
 * converts it a field-formatted text output file (or dataset).
 * <p>
 * <dt>Arguments to main:</dt>
 * <dd> inputfile   [outputfile] </dd>
 * <p>
 * <dt>Where:</dt>
 * <dt>inputfile</dt>
 * <dd>is the XML input file.  Can be  "//dataset.name", "//DD:ddname", or a 
 * normal java.io filepath.
 * 
 * <dt>outputfile</dt
 * <dd>is the output file.  Can be  "//dataset.name", "//DD:ddname", or a 
 * normal java.io filepath.  If not specified, defaults to java System.out
 * </dd>
 * 
 * @see Digester
 */
public class ParseXmlToFile {

	private String inputFilename;
	private String outputFilename;
	private BufferedWriter writer;
	private String padding;
	private int count = 0;

	/**
	 * The main method
	 */
	public static void main(String[] args) throws Exception {

		if (args.length < 1) {
			throw new IllegalArgumentException("arguments required: inputfile [outputfile]");
		}
		String inputFilename = args[0];
		String outputFilename = null;
		if (args.length >= 2) {
			outputFilename = args[1];
		}

		ParseXmlToFile instance = new ParseXmlToFile(
				inputFilename, outputFilename);
		instance.run();
	}

	public ParseXmlToFile(String inputFilename, String outputFilename) {

		this.inputFilename = inputFilename;
		this.outputFilename = outputFilename;
		
		// initialize a blank string used for padding fields
		char[] ca = new char[1000];
		for (int i=0; i<ca.length; i++) ca[i] = ' ';
		padding = new String(ca);
	}

	public void run() throws Exception {

		if (outputFilename == null) {
			writer = new BufferedWriter(new OutputStreamWriter(System.out));
		} else {
			writer = FileFactory.newBufferedWriter(outputFilename);
		}
		
		InputStream inputStream = null;
		
		try {
			
			// We open the input XML file as a raw unencoded inputstream
			// so that the XML parser can detect the proper encoding from
			// its prologue.
			inputStream = FileFactory.newInputStream(inputFilename);
			
			Digester digester = new Digester();
			
			// make the receiver's methods available to digester callback rules
			digester.push(this);
			
			// Define a callback method to handle each congressional member
			digester.addCallMethod("congress/member", "handleCongressMember", 7);
			digester.addCallParam("congress/member/district", 0);
			digester.addCallParam("congress/member/name", 1);
			digester.addCallParam("congress/member/party", 2);
			digester.addCallParam("congress/member/contact/address", 3);
			digester.addCallParam("congress/member/contact/phone", 4);
			digester.addCallParam("congress/member/contact/district-phone", 5);
			digester.addCallParam("congress/member/contact/url", 6);
			
			digester.parse(inputStream);
			
			System.out.println("" + count + " records written to output file");
		} finally {
			
			try {
			if (inputStream != null) inputStream.close();
			if (writer != null ) writer.close();
			} catch (Exception ignore) {}  // ignore close errors, they might have already been closed
		}
	}
	
	/**
	 * This is a callback method for the Digester where we write output records
	 */
	public void handleCongressMember (
			String district,
			String name, 
			String party,
			String address,
			String phone,
			String districtPhone,
			String url )  throws IOException {
		
		writer.write(getField(district, 	5));
		writer.write(getField(name, 		30));
		writer.write(getField(party, 		2));
		writer.write(getField(address, 		40));
		writer.write(getField(phone, 		13));
		writer.write(getField(districtPhone, 13));
		writer.write(getField(url, 			40));
		writer.newLine();
		count++;
	}
	
	/**
	 * A helper method which pads or truncates a field to a specified length 
	 */
	private String getField(String arg, int len) {
		
		if (arg == null) {
			return padding.substring(0,len);
		}
		int argLen = arg.length();
		if (argLen > len) return arg.substring(0,len);
		if (argLen == len) return arg;
		
		return arg + padding.substring(0, len - argLen);
	}

}
